
#include <iostream>
#include <fstream>
#include <sstream>
#include <memory>
#include <string>

#include <grpcpp/grpcpp.h>
#include "turbo.grpc.pb.h"

using namespace std;
using namespace grpc;
using namespace Turbo;


// types 

struct Option_t
{
	string name;
	string value;
};



// typedef for a function pointer like main
using MainLikeFuncPtr_t = int(*)(int argc, char *argv[], const vector<Option_t>& upper_opts) ;

struct SubCommand_t
{
	string            name;
	MainLikeFuncPtr_t func;


};

// would be nice to have in a header, but don't put odd macros in headers
#define ALLOF(a) begin(a),end(a)


string getOption(const vector<Option_t>& opts, const string& name)
{

	const auto found_it = find_if(ALLOF(opts), [&](const Option_t& opt){ return opt.name==name; });
	assert(found_it!=end(opts));
	return found_it->value;

}

int32_t getOptionInt(const vector<Option_t>& opts, const string& name)
{
	return atoi(getOption(opts,name).c_str());
}


// helper functions 
void print_error(const Status& status) 
{
	cout << "Server returned error. " << endl;
	cout << "Number:  " << status.error_code()    << endl; 
	cout << "Message: " << status.error_message() << endl;
	cout << "Details: " << status.error_details() << endl;
}

int Usage(int argc, char* argv[], const vector<Option_t>& upper_opts)
{
	cout << getOption(upper_opts, "pgm-name") << ":" << endl;
	cout << "	--help|-h|--usage|-?               Print this screen"                       << endl;
	cout << "	ping                               Verify that the Turbo server is running" << endl;
	cout << "	boost add <name>                   Create a new boost" << endl;
	cout << "	boost list                         List all existing boosts"<< endl;
	cout << "	boost show <id>                    Show details of a boost" << endl;
	cout << "	boost delete idname>               Delete an existing boost" << endl;
	cout << "	version add <boost id> <file>      Create a new version of a boost" << endl;
	cout << "	version list                       List all existing version of a boost"<< endl;
	cout << "	version show <id>                  Show details of a version of a boost" << endl;
	cout << "	version delete <id>                Delete an existing version of a boost" << endl;
	return 0;
}



shared_ptr<ChannelCredentials> getCredentials(const vector<Option_t>& opts)
{
	return grpc::InsecureChannelCredentials();
}

int subcommandDispatch(string cmd_msg, const vector<SubCommand_t> cmds, int argc, char* argv[], vector<Option_t> upper_opts)
{
	if(argc < 2)
	{
		cout << "Error:  must specify " << cmd_msg << " subcommand "<< endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	const auto subcommand_it = find_if(ALLOF(cmds), 
			[argv](const SubCommand_t& cmd) { return cmd.name == argv[1]; });
	if(subcommand_it ==  end(cmds))
	{
		cout << "Error:  Invalid " << cmd_msg << " subcommand:" << argv[1] << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	return subcommand_it->func(argc-1,argv+1, upper_opts);
}

//
// how to actually parse and perform a (sub)command
// 


// check server alive.
int Ping(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	const auto opt_val = getOption(upper_opts, "host");
	auto creds         = getCredentials(upper_opts);
	auto channel       = grpc::CreateChannel(opt_val,creds);
	auto backend       = TurboService::NewStub(channel); 

	// Data and answer we are exchanging with the server.
	auto request = PingRequest_t();
	auto reply   = PingReply_t();
	ClientContext context; // cannot auto here due to constructor/copy restrictions

	// perform the communication
	const auto status = backend->Ping(&context, request, &reply);

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		return 0;
	} 
	else 
	{
		print_error(status);
		return 1;
	}
}

int VersionAdd(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<3)
	{
		cout << "Error, version add must specify a file name and project id" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>3)
	{
		cout << "Error, junk at end of version add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	const auto opt_val = getOption(upper_opts, "host");
	const auto creds   = getCredentials(upper_opts);
	auto channel       = grpc::CreateChannel(opt_val,creds);
	auto backend       = TurboService::NewStub(channel); 

	auto request = VersionAddRequest_t();
	auto reply   = VersionAddReply_t();
	ClientContext context; 

	request.set_boost_id(atoi(argv[1]));

	ifstream fin(argv[2], ios::binary);
	if(!fin)
	{
		cout << "Cannot open file: " << argv[2] << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}

	ostringstream ostrm;
	ostrm << fin.rdbuf();
	const auto &send_contents = ostrm.str();

	request.set_version_contents(send_contents);

	const auto status = backend->VersionAdd(&context, request, &reply);

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		cout << "Version ID: " << reply.version_id() << endl;
		return 0;
	} 
	else 
	{
		print_error(status);
		return 1;
	}
}

int VersionList(int argc, char *argv[], const vector<Option_t>& upper_opt) 
{
	assert(0);
	return 1;
}

int VersionShow(int argc, char *argv[], const vector<Option_t>& upper_opt) 
{
	assert(0);
	return 1;
}

int VersionDelete(int argc, char *argv[], const vector<Option_t>& upper_opt) 
{
	assert(0);
	return 1;
}



//
// process the boost subcommand
//
int Version(int argc, char *argv[], const vector<Option_t>& upper_opt) 
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"add",    VersionAdd    },
			{"list",   VersionList   },
			{"show",   VersionShow   },
			{"delete", VersionDelete },
		};

	return subcommandDispatch("version", subcommands, argc, argv, upper_opt);
}


//
// add a "boost" to to the server
// 
int BoostAdd(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	if(argc<2)
	{
		cout << "Error, boost add must specify a boost name" << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	if(argc>2)
	{
		cout << "Error, junk at end of boost add " << endl;
		Usage(argc,argv, upper_opts);
		return 1;
	}
	const auto opt_val = getOption(upper_opts, "host");
	const auto creds   = getCredentials(upper_opts);
	auto channel       = grpc::CreateChannel(opt_val,creds);
	auto backend       = TurboService::NewStub(channel); 

	auto request = BoostAddRequest_t();
	auto reply   = BoostAddReply_t();
	ClientContext context; 

	request.set_name(argv[1]);

	const auto status = backend->BoostAdd(&context, request, &reply);

	// Act upon its status.
	if (status.ok()) 
	{
		cout << "Server returned OK" << endl;
		cout << "Boost ID: " << reply.boost_id() << endl;
		return 0;
	}
	else 
	{
		print_error(status);
		return 1;
	}
}


// 
// list all boosts
//
int BoostList(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

// 
// show a particular boost
//
int BoostShow(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

//
// remove a boost from the system
//
int BoostDelete(int argc, char *argv[], const vector<Option_t>& upper_opts) 
{
	assert(0);
	return 1;
}

//
// process the boost subcommand
//
int Boost(int argc, char *argv[], const vector<Option_t>& upper_opt) 
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"add",    BoostAdd    },
			{"list",   BoostList   },
			{"show",   BoostShow   },
			{"delete", BoostDelete },
		};

	return subcommandDispatch("boost", subcommands, argc, argv, upper_opt);
}


// 
// handle processing of the turbo cli
//
int main(int argc, char *argv[])
{
	const auto subcommands = vector<SubCommand_t> 
		{
			{"ping",    Ping    },
			{"boost",   Boost   },
			{"version", Version },
			{"--help",  Usage   },
			{"--usage", Usage   },
			{"-h",      Usage   },
			{"-?",      Usage   },
		};
	return subcommandDispatch("turbo-cli", subcommands, argc, argv, {{"pgm-name", argv[0]}, {"host", "localhost:55155"}});
}
